from sqlalchemy import ForeignKey
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

from nextline_rdb.models import ReprMixin


class Model(ReprMixin, DeclarativeBase):
    pass


class Foo(Model):
    __tablename__ = 'foo'
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    bars: Mapped[list['Bar']] = relationship(back_populates='foo')


class Bar(Model):
    __tablename__ = 'bar'
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    foo_id: Mapped[int] = mapped_column(ForeignKey('foo.id'))
    foo: Mapped[Foo] = relationship(back_populates='bars')



from logging import getLogger
from typing import Optional

from nextline import Nextline
from sqlalchemy import func, select
from sqlalchemy.ext.asyncio import AsyncSession

from . import models
from .db import DB


async def initialize_nextline(nextline: Nextline, db: DB) -> None:
    run_no, script = await _last_run_no_and_script(db)
    if run_no is not None:
        run_no += 1
        if run_no >= nextline._init_options.run_no_start_from:
            nextline._init_options.run_no_start_from = run_no
    if script is not None:
        nextline._init_options.statement = script


async def _last_run_no_and_script(db: DB) -> tuple[Optional[int], Optional[str]]:
    async with db.session() as session:
        last_run = await _last_run(session)
        if last_run is None:
            return None, None
        else:
            return last_run.run_no, last_run.script


async def _last_run(session: AsyncSession) -> Optional[models.Run]:
    stmt = select(models.Run, func.max(models.Run.run_no))
    if model := (await session.execute(stmt)).scalar_one_or_none():
        return model
    else:
        logger = getLogger(__name__)
        msg = "No previous runs were found in the DB"
        logger.info(msg)
        return None



from typing import NamedTuple, Optional, Type, TypeVar, cast

from sqlalchemy import func, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeBase, aliased
from sqlalchemy.sql.expression import literal
from sqlalchemy.sql.selectable import Select

# import sqlparse


# def format_sql(sql):
#     return sqlparse.format(str(sql), reindent=True, keyword_case='upper')


class SortField(NamedTuple):
    field: str
    desc: bool = False


Sort = list[SortField]

_Id = TypeVar('_Id')


async def load_models(
    session: AsyncSession,
    Model: Type[DeclarativeBase],
    id_field: str,
    *,
    sort: Optional[Sort] = None,
    before: Optional[_Id] = None,
    after: Optional[_Id] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
):
    stmt = compose_statement(
        Model,
        id_field,
        sort=sort,
        before=before,
        after=after,
        first=first,
        last=last,
    )

    models = await session.scalars(stmt)
    return models


def compose_statement(
    Model: Type[DeclarativeBase],
    id_field: str,
    *,
    sort: Optional[Sort] = None,
    before: Optional[_Id] = None,
    after: Optional[_Id] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Select:
    '''Return a SELECT statement object to be given to session.scalars'''

    forward = (after is not None) or (first is not None)
    backward = (before is not None) or (last is not None)

    if forward and backward:
        raise ValueError('Only either after/first or before/last is allowed')

    sort = sort or []

    if id_field not in [s.field for s in sort]:
        sort.append(SortField(id_field))

    def sorting_fields(Model, reverse=False):
        return [
            f.desc() if reverse ^ d else f
            for f, d in [(getattr(Model, s.field), s.desc) for s in sort]
        ]

    if not (forward or backward):
        return select(Model).order_by(*sorting_fields(Model))

    cursor = after if forward else before
    limit = first if forward else last

    if cursor is None:
        stmt = select(Model).order_by(*sorting_fields(Model, reverse=backward))
    else:
        cte = select(
            Model,
            func.row_number()
            .over(order_by=sorting_fields(Model, reverse=backward))
            .label('row_number'),
        ).cte()

        subq = select(cte.c.row_number.label('cursor'))
        subq = subq.where(getattr(cte.c, id_field) == cursor)
        subq = cast(Select[tuple], subq.subquery())

        Alias = aliased(Model, cte)  # type: ignore
        stmt = select(Alias).select_from(cte)
        stmt = stmt.join(subq, literal(True))  # type: ignore # cartesian product
        stmt = stmt.order_by(*sorting_fields(Alias, reverse=backward))
        stmt = stmt.where(cte.c.row_number > subq.c.cursor)

    if limit is not None:
        stmt = stmt.limit(limit)

    if backward:
        Alias = aliased(Model, stmt.subquery())
        stmt = select(Alias).order_by(*sorting_fields(Alias))

    return stmt
